#brailleDisplayDrivers/braillex.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright (C) 2010 Gianluca Casalino <gianluca@spazioausili.net>

import time
import wx
import braille
from logHandler import log
from keyboardHandler import KeyboardInputGesture
import os
devices = {'55': {'Name': 'BRAILLEX EL40s', 'vCells': 0, 'hCells': 40, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'A'}, \
	'56': {'Name': 'BRAILLEX EL80 II', 'vCells': 0, 'hCells': 82, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'A'}, \
	'57': {'Name': 'BRAILLEX EL66s', 'vCells': 0, 'hCells': 66, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'A'}, \
	'58': {'Name': 'BRAILLEX EL80', 'vCells': 0, 'hCells': 80, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'A'}, \
	'59': {'Name': 'BRAILLEX Trio', 'vCells': 0, 'hCells': 40, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'B'}, \
	'5A': {'Name': 'BRAILLEX EL70s', 'vCells': 0, 'hCells': 70, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'A'}, \
	'5B': {'Name': 'BRAILLEX EL2D-80s', 'vCells': 40, 'hCells': 80, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'A'}, \
	'5C': {'Name': 'BRAILLEX Elba', 'vCells': 0, 'hCells': 20, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'B'}, \
	'5D': {'Name': 'BRAILLEX Elba', 'vCells': 0, 'hCells': 32, 'nlk': 1, 'nrk': 1, 'protocolRevision': 'B'}}

FHP_ID = 0x403F208
STX = '\x02'
ETX = '\x03'
CMD_BRAILLE_OUTPUT = '\x43'
REQUEST_AUTO_ID_SEQUENCE = STX+'\x42\x50\x50'+ETX
KEY_CHECK_INTERVAL = 50 

#MAP OF KEYS

KEY_LEFT_UP = 24
KEY_LEFT_DOWN = 25
KEY_RIGHT_UP = 10
KEY_RIGHT_DOWN = 11
BAR_L2_UP = 16
BAR_L1_UP = 17
BAR_L1_DOWN = 18
BAR_L2_DOWN = 19
BAR_L2_LEFT = 23
BAR_L1_LEFT = 21
BAR_L1_RIGHT = 20
BAR_L2_RIGHT = 22

try:
	import ftdi
except ImportError:
 	pass

def findPmDevice():
	""" search if any ftdi devices are connected and checks if one of them is a Papenmeier device returning its Serial Number
	"""  
	devicesList = ftdi.list_devices()
	if not devicesList: return None
	ftdi.create_device_info_list()
	for x in xrange(len(devicesList)): 
		deviceInfo = ftdi.get_device_info_detail(x)
		if deviceInfo['ID'] == FHP_ID: return deviceInfo['SerialNumber']
	return None

def convertCell(cell):
	""" dots cell mapping
	"""
	return ((1<<7 if cell & 1<<0 else 0) | 
	(1<<6 if cell & 1<<1 else 0) | 
	(1<<5 if cell & 1<<2 else 0) | 
	(1<<4 if cell & 1<<3 else 0) | 
	(1<<3 if cell & 1<<4 else 0) | 
	(1<<2 if cell & 1<<5 else 0) | 
	(1<<1 if cell & 1<<6 else 0) | 
	(1<<0 if cell & 1<<7 else 0))

class BrailleDisplayDriver(braille.BrailleDisplayDriverWithCursor):
	"""Papenmeier Braillex II braille display driver family.
	"""
	name = "braillex"
	description = "Papenmeier Braillex EL-S/Trio USB devices"

	@classmethod
	def check(cls):
		if isinstance(braille.handler.display, cls):
			return True
		return bool(findPmDevice())

	def __init__(self):
		super(BrailleDisplayDriver, self).__init__()
		self.prevKeyTrame = '\x30'*33
		self.keys = []
		self.cellsMap = [convertCell(x) for x in range(256)] 
		if not bool(findPmDevice()): raise RuntimeError("Papenmeier device not found")
		try:
			self.fhpHandle = ftdi.open_ex(findPmDevice())
			self.fhpHandle.set_baud_rate(115200)
			self.fhpHandle.purge()
			self.fhpHandle.write(REQUEST_AUTO_ID_SEQUENCE)
			if self.fhpHandle.get_queue_status() == 0: 
				self.fhpHandle.set_baud_rate(57600)
				self.fhpHandle.write(REQUEST_AUTO_ID_SEQUENCE)
		except:
			raise RuntimeError("Papenmeier device not found")
		devIdCode = self.fhpHandle.read(10)[4:6]  
		if devIdCode == '': raise RuntimeError("Papenmeier device not found")
		self.dev = devices[devIdCode]
		if self.dev['protocolRevision'] == 'B':
			self.sequenceLength = self.dev['vCells']+self.dev['hCells']
		else:
			self.sequenceLength = self.dev['vCells']+self.dev['hCells']+self.dev['nlk']*2+self.dev['nrk']*2
		self.keyPacketLength = ((5+2*self.dev['nlk']+2*self.dev['nrk']+2+ 
		2*(self.dev['vCells']*2)/8+2*(self.dev['nrk']*4)/8+ 
		2*(self.dev['hCells']*2)/8+2*(self.dev['nrk']*4)/8))
		self._keyCheckTimer = wx.PyTimer(self._handleKeyPresses)
		self._keyCheckTimer.Start(KEY_CHECK_INTERVAL)

	def terminate(self):
		super(BrailleDisplayDriver, self).terminate()
		try:
			self._keyCheckTimer.Stop()
			self._keyCheckTimer = None
			self.fhpHandle.close()
			del self.fhpHandle
		except:
			pass

 	def _get_numCells(self):
		return self.dev['hCells']

	def _display(self, cells):
		cells = "".join([chr(0x30 | self.cellsMap[x]>>4)+chr(0x30 | self.cellsMap[x] & 0x0f) for x in cells]) 
		sequence = STX+CMD_BRAILLE_OUTPUT
		sequence += chr(0x50 | self.sequenceLength >> 4)+chr(0x50 | self.sequenceLength & 0x0F)
		if self.dev['protocolRevision'] != 'B': sequence+='\x30'*(2*(2*self.dev['nlk']))
		sequence += '\x30'*(2*self.dev['vCells'])+cells
		if self.dev['protocolRevision'] != 'B': sequence+='\x30'*(2*(2*self.dev['nrk']))
		sequence += ETX
		self.fhpHandle.write(sequence) 

	def _handleKeyPresses(self):
		if self.dev['protocolRevision'] == 'B': return
		if self.fhpHandle.get_queue_status() >= self.keyPacketLength:
			trame = self.fhpHandle.read(self.fhpHandle.get_queue_status())
			if trame[-1] != ETX: 
				while trame[-1] != ETX: trame += self.fhpHandle.read(self.fhpHandle.get_queue_status())
			self.readTrame(trame)
		if self.keys: 
			for key in range(len(self.keys)): self._onKeyPress(self.keys.pop()) 

	def _onKeyPress(self, keyPressed):
		if keyPressed == BAR_L1_LEFT: braille.handler.scrollBack()
		elif keyPressed == BAR_L1_RIGHT: braille.handler.scrollForward()
		elif keyPressed >= 28 < (28+(self.dev['hCells']*2)): braille.handler.routeTo((keyPressed-28)/2)
		elif keyPressed == BAR_L1_UP: KeyboardInputGesture.fromName('upArrow').send()
		elif keyPressed == BAR_L1_DOWN: KeyboardInputGesture.fromName('downArrow').send()
		elif keyPressed == BAR_L2_LEFT: KeyboardInputGesture.fromName('shift+tab').send()
		elif keyPressed == BAR_L1_RIGHT: KeyboardInputGesture.fromName('downArrow').send()
		elif keyPressed == BAR_L2_RIGHT: KeyboardInputGesture.fromName('tab').send()
		elif keyPressed == KEY_LEFT_UP: KeyboardInputGesture.fromName('alt+f4').send()
		elif keyPressed == KEY_LEFT_DOWN: KeyboardInputGesture.fromName('alt+tab').send()
		elif keyPressed == BAR_L2_UP: KeyboardInputGesture.fromName('home').send()
		elif keyPressed == BAR_L2_DOWN: KeyboardInputGesture.fromName('end').send()
		elif keyPressed == KEY_RIGHT_UP: KeyboardInputGesture.fromName('win+d').send()
		elif keyPressed == KEY_RIGHT_DOWN: KeyboardInputGesture.fromName('win').send()

	def readTrame(self,trame):
		""" passes the buffer recheived from the device through splitting it in valid packets and processing them if key pressed packet. a list of key values is returned
		"""
		packets = trame.split(STX)
		for keyTrame in packets: 
			if ((len(keyTrame) == self.keyPacketLength-1) and (keyTrame[0:3] == '\x4b'+chr(0x50 | (self.keyPacketLength-5)/2>>4)+chr(0x50 | (self.keyPacketLength-5)/2 & 0x0f) ) and (keyTrame[-1] == '\x03')): 
				keyTrame = keyTrame[3:-1]
				for x in xrange(len(keyTrame)):
					byte = (ord(keyTrame[x]) & 0x0F if  (x % 2) else (ord(keyTrame[x]) & 0x0F)<<4) 
					bytePrev = (ord(self.prevKeyTrame[x]) & 0x0F if  (x % 2) else (ord(self.prevKeyTrame[x]) & 0x0F)<<4) 
					for j in xrange(8): 
						if   ((byte & 1<<j) and  not (bytePrev & 1<<j)) and not  (not (byte & 1<<j) and (bytePrev & 1<<j)): 
							bit = j
							key = (x/2*8+bit)-self.dev['vCells']*2
							if key == (28+(self.dev['hCells']*2)): key=10
							elif key == 28+(self.dev['hCells']*2)+1: key=11
							self.keys.append(key)
				self.prevKeyTrame = keyTrame
